import cppyy
import numpy as np
from .ortho_tools import RPCCoeffs

cpp_source = """
#include <cmath>


template <class T> constexpr float interpolate(float x, float y, T source, int source_height) {
    int x_floor = static_cast<int>(std::floor(x));
    int x_ceil =  static_cast<int>(std::ceil(x));
    int y_floor =  static_cast<int>(std::floor(y));
    int y_ceil =  static_cast<int>(std::ceil(y));

    float x_frac = x - x_floor;
    float y_frac = y - y_floor;
    
    long ul_index = x_floor + source_height * y_floor;
    long ur_index = x_ceil + source_height * y_floor;
    long lr_index = x_ceil + source_height * y_ceil;
    long ll_index = x_floor + source_height * y_ceil;
    
    float ul = static_cast<float>(source[ul_index]);
    float ur = static_cast<float>(source[ur_index]);
    float lr = static_cast<float>(source[lr_index]);
    float ll = static_cast<float>(source[ll_index]);
    
    float upper_x = (1 - x_frac) * ul + x_frac * ur;
    float lower_x = (1 - x_frac) * ll + x_frac * lr;
    float upper_y = (1 - y_frac) * ul + y_frac * ll;
    float lower_y = (1 - y_frac) * ur + y_frac * lr;
    float final_value = (upper_x + lower_x + upper_y + lower_y) / 4;
    return final_value;
}

void lon_lat_alt_to_xy_cpp(
    float lon,
    float lat,
    float alt,
    float height_off,
    float height_scale,
    float lat_off,
    float lat_scale,
    float* line_den_coeff,
    float* line_num_coeff,
    float line_off,
    float line_scale,
    float long_off,
    float long_scale,
    float max_lat,
    float max_long,
    float min_lat,
    float min_long,
    float* samp_den_coeff,
    float* samp_num_coeff,
    float samp_off,
    float samp_scale,
    float* output
) {
    
    float norm_lon = (lon - long_off) / long_scale;
    float norm_lat = (lat - lat_off) / lat_scale;
    float norm_alt = (alt - height_off) / height_scale;
    
    float formula[20] = {
        1,
        norm_lon,
        norm_lat,
        norm_alt,
        norm_lon * norm_lat,
        norm_lon * norm_alt,
        norm_lat * norm_alt,
        static_cast<float>(std::pow(norm_lon, 2)),
        static_cast<float>(std::pow(norm_lat, 2)),
        static_cast<float>(std::pow(norm_alt, 2)),
        norm_lat * norm_lon * norm_alt,
        static_cast<float>(std::pow(norm_lon, 3)),
        norm_lon * static_cast<float>(std::pow(norm_lat, 2)),
        norm_lon * static_cast<float>(std::pow(norm_alt, 2)),
        static_cast<float>(std::pow(norm_lon, 2)) * norm_lat,
        static_cast<float>(std::pow(norm_lat, 3)),
        norm_lat * static_cast<float>(std::pow(norm_alt, 2)),
        static_cast<float>(std::pow(norm_lon, 2)) * norm_alt,
        static_cast<float>(std::pow(norm_lat, 2)) * norm_alt,
        static_cast<float>(std::pow(norm_alt, 3))
    };
    
    float f1 = 0.0;
    for (int i = 0; i < 20; i++) 
        f1 += samp_num_coeff[i] * formula[i];
    
    float f2 = 0.0;
    for (int i = 0; i < 20; i++) 
        f2 += samp_den_coeff[i] * formula[i];
       
    float f3 = 0.0;
    for (int i = 0; i < 20; i++) 
        f3 += line_num_coeff[i] * formula[i];
    
    float f4 = 0.0;
    for (int i = 0; i < 20; i++) 
        f4 += line_den_coeff[i] * formula[i];   

    float samp_number_normed = f1 / f2;
    float line_number_normed = f3 / f4;
    
    output[0] = samp_number_normed * samp_scale + samp_off;
    output[1] = line_number_normed * line_scale + line_off;

}

void make_ortho_cpp(
    float x1,
    float x2,
    float y1,
    float y2,
    int width,
    int height, 
    uint8_t* source,
    int source_width,
    int source_height,
    float height_off,
    float height_scale,
    float lat_off,
    float lat_scale,
    float* line_den_coeff,
    float* line_num_coeff,
    float line_off,
    float line_scale,
    float long_off,
    float long_scale,
    float max_lat,
    float max_long,
    float min_lat,
    float min_long,
    float* samp_den_coeff,
    float* samp_num_coeff,
    float samp_off,
    float samp_scale,
    float* dem,
    int dem_width,
    int dem_height,
    float* dem_geot,
    uint8_t* output
) {

    float gsd_x = abs(x2 - x1) / width;
    float gsd_y = abs(y2 - y1) / height;
    for (int i = 0; i < width; i++) {
        float lon = x1 + ((float) i * gsd_x);
        for (int j = 0; j < height; j++) {
            float lat = y1 + ((float) j * gsd_y);
            float approx_dem_x = ((lon - dem_geot[0]) / dem_geot[1]);
            float approx_dem_y = ((lat - dem_geot[3]) / dem_geot[5]);    
            float altitude = interpolate(approx_dem_x, approx_dem_y, dem, dem_height);
            
            float* xy_loc = (float*) malloc(sizeof(float) * 2);
            lon_lat_alt_to_xy_cpp(
                lon,
                lat,
                altitude,
                height_off,
                height_scale,
                lat_off,
                lat_scale,
                line_den_coeff,
                line_num_coeff,
                line_off,
                line_scale,
                long_off,
                long_scale,
                max_lat,
                max_long,
                min_lat,
                min_long,
                samp_den_coeff,
                samp_num_coeff,
                samp_off,
                samp_scale,
                xy_loc
            );
            
            float x = xy_loc[0];
            float y = xy_loc[1];
            free(xy_loc);

            if (x > 1 && x < source_height - 1 && y > 1 && y < source_width - 1) {
                uint8_t result = static_cast<uint8_t>(
                    std::round(interpolate(x, y, source, source_height))
                );
                output[i * height + j] = result;
            }
        }
    }
}
"""

cppyy.cppdef(cpp_source)
from cppyy.gbl import make_ortho_cpp


def ortho_cpp(
    x1: float,
    x2: float,
    y1: float,
    y2: float,
    width: int,
    rpcs: RPCCoeffs,
    source: np.ndarray,
    dem: np.ndarray,
    dem_geot: np.array,
):
    gsd = abs(x1 - x2) / width
    height = int(abs(y1 - y2) / gsd)
    output = np.zeros((width * height), dtype=np.ubyte)
    make_ortho_cpp(
        x1,
        x2,
        y1,
        y2,
        width,
        height,
        source.reshape(-1),
        source.shape[0],
        source.shape[1],
        *rpcs,
        dem.astype(np.float32).reshape(-1),
        dem.shape[0],
        dem.shape[1],
        dem_geot.astype(np.float32),
        output
    )
    return output.reshape(width, height).transpose(), gsd, x1, y1



import numpy as np
import math
from sardem.dem import main as load_dem
from numba import jit, prange
from typing import Tuple, NamedTuple
import gdal


class RPCCoeffs(NamedTuple):
    height_off: float
    height_scale: float
    lat_off: float
    lat_scale: float
    line_den_coeff: np.array
    line_num_coeff: np.array
    line_off: float
    line_scale: float
    long_off: float
    long_scale: float
    max_lat: float
    max_long: float
    min_lat: float
    min_long: float
    samp_den_coeff: np.array
    samp_num_coeff: np.array
    samp_off: float
    samp_scale: float


@jit(nopython=True)
def lon_lat_alt_to_xy(
    lon: float,
    lat: float,
    alt: float,
    rpcs: RPCCoeffs,
) -> Tuple[float, float]:
    """
    Returns an image pixel coordinate (x, y) corresponding to a provided world coordinate (lon, lat, alt)
    using provided RPC coefficients
    :param lon: The world coordinate longitude
    :param lat: The world coordinate latitude
    :param alt: The world coordinate altitude
    :param rpcs: NamedTuple with RPC coefficients
    :param interp: Interpolation method -- Linear and Nearest are supported
    :return: Pixel coordinate (x,y) as floating point values -- interpolation will be needed to arrive at a
    pixel intensity value!
    """
    # First create the normalized values for lon/lat/alt
    norm_lon = (lon - rpcs.long_off) / rpcs.long_scale
    norm_lat = (lat - rpcs.lat_off) / rpcs.lat_scale
    norm_alt = (alt - rpcs.height_off) / rpcs.height_scale

    # Create the polynomial vector (gets re-used)
    formula = np.array(
        [
            1,
            norm_lon,
            norm_lat,
            norm_alt,
            norm_lon * norm_lat,
            norm_lon * norm_alt,
            norm_lat * norm_alt,
            norm_lon ** 2,
            norm_lat ** 2,
            norm_alt ** 2,
            norm_lat * norm_lon * norm_alt,
            norm_lon ** 3,
            norm_lon * (norm_lat ** 2),
            norm_lon * (norm_alt ** 2),
            (norm_lon ** 2) * norm_lat,
            norm_lat ** 3,
            norm_lat * (norm_alt ** 2),
            (norm_lon ** 2) * norm_alt,
            (norm_lat ** 2) * norm_alt,
            norm_alt ** 3,
        ],
        dtype=np.float32
    )

    # PLUG AND CHUG
    f1 = np.dot(rpcs.samp_num_coeff, formula)
    f2 = np.dot(rpcs.samp_den_coeff, formula)
    f3 = np.dot(rpcs.line_num_coeff, formula)
    f4 = np.dot(rpcs.line_den_coeff, formula)

    samp_number_normed = f1 / f2
    line_number_normed = f3 / f4

    # Then denormalize to get the approximate pixel coordinate
    samp_number = samp_number_normed * rpcs.samp_scale + rpcs.samp_off
    line_number = line_number_normed * rpcs.line_scale + rpcs.line_off

    return samp_number, line_number


@jit(nopython=True, parallel=True, nogil=True)
def make_ortho(
    x1: float,
    x2: float,
    y1: float,
    y2: float,
    width: int,
    source: np.ndarray,
    rpcs: RPCCoeffs,
    dem: np.ndarray,
    dem_geot: np.array,
) -> Tuple[np.array, float, float, float]:
    """
    Produces an orthorectified image given a
    :param x1: upper left x
    :param x2: lower right x
    :param y1: upper left y
    :param y2: lower right y
    :param width: num pixels for the width of the resulting orthorectified image (height will be set by aspect ratio)
    :param source: the original raw image
    :param rpcs: rational polynomial coefficients object
    :param dem: dem tile corresponding to the image location
    :param dem_geot: dem geo transform (affine transform)
    :param interp: Interpolation type, 0 == nearest neighbor, 1 == bilinear
    :return: a Tuple containing the ortho'd image, the ground sampling distance in degrees,
    and the upper left coordinate
    """
    cols = np.linspace(x1, x2, width)
    gsd = abs(cols[0] - cols[1])
    height = int(abs(y1 - y2) / gsd)
    rows = np.linspace(y1, y2, height)
    ortho = np.zeros(width * height)
    for i in prange(len(cols)):
        lon = cols[i]
        for j in prange(len(rows)):
            lat = rows[j]
            dem_x = int((lon - dem_geot[0]) / dem_geot[1])
            dem_y = int((lat - dem_geot[3]) / dem_geot[5])
            if dem_x < 0 or dem_x > dem.shape[1] - 1 or dem_y < 0 or dem_y > dem.shape[0] - 1:
                # Numba will segfault if I don't catch this... AND it will fail to compile
                # if I try to include a useful message with "dem_x" and "dem_y" :( :( :(
                raise IndexError("DEM indices out of bounds")
            altitude = linear_interp(dem_x, dem_y, dem.reshape(-1), dem.shape[1])
            x, y = lon_lat_alt_to_xy(lon, lat, altitude, rpcs)
            if 1 <= x < source.shape[1] - 1 and 1 <= y < source.shape[0] - 1:
                idx = i * height + j
                result = linear_interp(x, y, source.reshape(-1), source.shape[1])
                ortho[idx] = result
            else:
                pass
    return ortho.reshape(width, height).transpose(), gsd, x1, y1


@jit(nopython=True)
def linear_interp(x: float, y: float, source: np.ndarray, source_height: int) -> int:
    x_floor = math.floor(x)
    x_ceil = math.ceil(x)
    y_floor = math.floor(y)
    y_ceil = math.ceil(y)

    x_frac = x - x_floor
    y_frac = y - y_floor

    ul_index = x_floor + source_height * y_floor
    ur_index = x_ceil + source_height * y_floor
    lr_index = x_ceil + source_height * y_ceil
    ll_index = x_floor + source_height * y_ceil

    ul = source[ul_index]
    ur = source[ur_index]
    lr = source[lr_index]
    ll = source[ll_index]

    upper_x = (1 - x_frac) * ul + x_frac * ur
    lower_x = (1 - x_frac) * ll + x_frac * lr
    upper_y = (1 - y_frac) * ul + y_frac * ll
    lower_y = (1 - y_frac) * ur + y_frac * lr
    final_value = (upper_x + lower_x + upper_y + lower_y) / 4

    return int(final_value)


def unpack_rpc_parameters(dataset: gdal.Dataset) -> RPCCoeffs:
    """
    Returns RPC coefficients collection as a NamedTuple
    when provided with a GDAL dataset if that dataset contains RPCs
    :param dataset: GDAL dataset reference for an image with RPCs
    :return: A NamedTuple containing RPC coefficients and parameters
    """
    rpc_dict = dataset.GetMetadata_Dict("RPC")
    height_off = float(rpc_dict["HEIGHT_OFF"])
    height_scale = float(rpc_dict["HEIGHT_SCALE"])
    lat_off = float(rpc_dict["LAT_OFF"])
    lat_scale = float(rpc_dict["LAT_SCALE"])
    line_den_coeff = np.array(
        [
            float(coeff.strip())
            for coeff in rpc_dict["LINE_DEN_COEFF"].strip().split(" ")
        ],
        dtype=np.float32
    )
    line_num_coeff = np.array(
        [
            float(coeff.strip())
            for coeff in rpc_dict["LINE_NUM_COEFF"].strip().split(" ")
        ],
        dtype=np.float32
    )
    line_off = float(rpc_dict["LINE_OFF"])
    line_scale = float(rpc_dict["LINE_SCALE"])
    long_off = float(rpc_dict["LONG_OFF"])
    long_scale = float(rpc_dict["LONG_SCALE"])
    max_lat = float(rpc_dict["MAX_LAT"])
    max_long = float(rpc_dict["MAX_LONG"])
    min_lat = float(rpc_dict["MIN_LAT"])
    min_long = float(rpc_dict["MIN_LONG"])
    samp_den_coeff = np.array(
        [
            float(coeff.strip())
            for coeff in rpc_dict["SAMP_DEN_COEFF"].strip().split(" ")
        ],
        dtype=np.float32
    )
    samp_num_coeff = np.array(
        [
            float(coeff.strip())
            for coeff in rpc_dict["SAMP_NUM_COEFF"].strip().split(" ")
        ],
        dtype=np.float32
    )
    samp_off = float(rpc_dict["SAMP_OFF"])
    samp_scale = float(rpc_dict["SAMP_SCALE"])

    return RPCCoeffs(
        height_off,
        height_scale,
        lat_off,
        lat_scale,
        line_den_coeff,
        line_num_coeff,
        line_off,
        line_scale,
        long_off,
        long_scale,
        max_lat,
        max_long,
        min_lat,
        min_long,
        samp_den_coeff,
        samp_num_coeff,
        samp_off,
        samp_scale,
    )


def lon_lat_to_pixel(lon: float, lat: float, geot: np.array) -> Tuple[float, float]:
    """
    Returns a pixel coordinate (x,y) as a tuple when provided with a longitude, latitude coordinate
    and a geo_transform (affine transform)
    :param lon: longitude of the world coordinate
    :param lat: latitude of the world coordinate
    :param geot: affine transform parameters in a length 6 np.array
    :return: A pixel coordinate as a Tuple (x, y)
    """
    x = (lon - geot[0]) / geot[1]
    y = (lat - geot[3]) / geot[5]
    return x, y


def retrieve_dem(
    min_lon: float,
    min_lat: float,
    degrees_lon: float,
    degrees_lat: float,
    sampling_rate: int = 1,
    output_path: str = "/tmp/elevation.dem",
) -> Tuple[np.ndarray, np.array]:
    """
    Load SRTM tiles for a bouding rectangle defined by an upper left point (min_long, min_lat) and a width and height
    in degrees. Optionally, an integer sampling rate greater than 1 may be passed in to downsample the DEM.
    :param min_lon: x component of the upper left corner of the bounding rectangle
    :param min_lat: y component of the upper left corner of the bounding rectangle
    :param degrees_lon: width of the DEM in degrees
    :param degrees_lat: height of the DEM in degrees
    :param sampling_rate: sampling rate
    :param output_path: Path where the dem file is saved
    :return: A numpy ndarray with elevation data, followed by the geotransform of the elevation data wrapped in a Tuple
    """
    load_dem(
        min_lon,
        min_lat,
        degrees_lon,
        degrees_lat,
        rate=sampling_rate,
        data_source="AWS",
        output_name=output_path,
    )
    gdal.AllRegister()
    dem_dataset = gdal.Open(output_path)
    elevation_data = dem_dataset.ReadAsArray()
    geo_transform = np.array([*dem_dataset.GetGeoTransform()])
    return elevation_data, geo_transform



import gdal
import osr
import numpy as np


def save_raster_as_geotiff(ortho: np.ndarray, ul_lon: float, ul_lat: float, gsd: float, filename: str) -> None:
    io_driver = gdal.GetDriverByName('GTiff')
    ortho_ds = io_driver.Create(filename, ortho.shape[1], ortho.shape[0], 1, gdal.GDT_Byte)
    ortho_ds.SetGeoTransform((ul_lon, gsd, 0, ul_lat, 0, gsd))
    ortho_band = ortho_ds.GetRasterBand(1)
    ortho_band.WriteArray(ortho)
    coordinate_system = osr.SpatialReference()
    coordinate_system.ImportFromEPSG(4326)
    ortho_ds.SetProjection(coordinate_system.ExportToWkt())
    ortho_band.FlushCache()
    # makes sure resources are released...
    ortho_band = None
    ortho_ds = None


import cv2
import numpy as np


def resize(img: np.array, factor: int) -> np.array:
    img_resized = cv2.resize(
        img, (int(img.shape[0] / factor), int(img.shape[1] / factor))
    )
    return img_resized


def convert_to_8bit(img: np.array) -> np.array:
    img_rescaled = img / np.amax(img) * 255.0
    img_rescaled = img_rescaled.astype(np.ubyte)
    return img_rescaled


def gaussian_rescale(img: np.array, bitness=11, stdev_bound=3) -> np.array:
    img_float = img.astype(np.float32)
    img_0_1 = img_float / ((2 ** bitness) - 1)
    mean = img_0_1.mean()
    stdev = img_0_1.std()
    img_8_bit = img_0_1 + (0.5 - mean)
    new_min, new_max = 0.5 - stdev_bound * stdev, 0.5 + stdev_bound * stdev
    img_8_bit = img_8_bit - new_min
    img_8_bit = img_8_bit / new_max
    img_8_bit = img_8_bit * 255
    img_8_bit[np.where(img_8_bit > 255)] = 255
    img_8_bit[np.where(img_8_bit < 0)] = 0
    return img_8_bit



import cv2
import numpy as np
from shapely.geometry import Polygon
from typing import Sequence, Tuple
from numba import jit
from .ortho_tools import lon_lat_to_pixel


@jit(nopython=True)
def fracture_polygon_north_up(poly: Polygon, factor_x: int, factor_y) -> Sequence[Polygon]:
    """
    Breaks a Shapely polygon into small north up rectangles, attempts to fill as much area as possible
    with smaller squares. The factors control the number of rectangles per dimension (resolution)
    :param poly:
    :param factor_x: width of the resulting grid
    :param factor_y: height of the resulting grid
    :return: Collection of small north up rectangles (as Shapely Polygons)
    """
    env = poly.envelope.exterior
    min_x = min([coord[0] for coord in env.coords])
    min_y = min([coord[1] for coord in env.coords])
    max_x = max([coord[0] for coord in env.coords])
    max_y = max([coord[1] for coord in env.coords])
    xs = np.linspace(min_x, max_x, factor_x)
    ys = np.linspace(min_y, max_y, factor_y)
    polygons = []
    for i in range(len(xs) - 1):
        for j in range(len(ys) - 1):
            cell = Polygon(
                [
                    [xs[i], ys[j]],
                    [xs[i + 1], ys[j]],
                    [xs[i + 1], ys[j + 1]],
                    [xs[i], ys[j + 1]],
                    [xs[i], ys[j]],
                ]
            )
            if cell.within(poly):
                polygons.append(cell)
    return polygons


@jit(nopython=True)
def fracture_parallelogram(poly: Polygon, factor: int) -> Sequence[Polygon]:
    """
    Fractures a parallelogram into smaller parallelogram with a non-regular orientation.
    The orientation of each individual small parallelogram with respect to the original
    parallelogram is maintained.
    :param poly: Shapely polygon that must be a parallelogram (quadrilateral with 2 pairs of parallel sides)
    :param factor: Number of times per dimension to fracture the parallelogram
    :return: Collection of small parallelograms as Shapely polygons. No verification of parallelogram membership
    is performed on the input polygon, so expect absolute junk if you provide a non-parallelogram polygon
    """
    quadrilateral = poly.exterior
    v1 = quadrilateral.coords[0]
    v_across = quadrilateral.coords[1]
    v_down = quadrilateral.coords[3]
    step_across = (v_across - v1) / factor
    step_down = (v_down - v1) / factor

    polygons = []
    for i in range(factor):
        for j in range(factor):
            ul = v1 + (i * step_across) + (j * step_down)
            ur = v1 + ((i + 1) * step_across) + (j * step_down)
            ll = v1 + ((i + 1) * step_across) + ((j + 1) * step_down)
            lr = v1 + (i * step_across) + ((j + 1) * step_down)
            cell = Polygon([ul, ur, lr, ll, ul])
            polygons.append(cell)
    return polygons


def rescale_elevation_data(elevation_data: np.ndarray) -> np.ndarray:
    """
    Maps elevation data into a 0-255 8-bit representation that is suitable for viewing
    :param elevation_data: elevation data input image
    :return: rescaled elevation image
    """
    rescaled_elevation_data = elevation_data - elevation_data.min()
    rescaled_elevation_data = (
        rescaled_elevation_data / rescaled_elevation_data.max() * 255
    )
    return rescaled_elevation_data


def reproject_with_affine(
    coords: Sequence[Tuple[float, float]],
    geo_transform: np.array,
    resize_factor: float = 1.0,
) -> Sequence[Tuple[float, float]]:
    """
    Reprojects a coordinate array according to a provided affine transform and resizing factor
    :param coords: Coordinate array that will be reprojected
    :param geo_transform: Affine transform as the basis of the reprojection  (length 6 np.array of float32/64)
    :param resize_factor: Optional resizing factor, useful if you want to display reprojected points on a downsampled
    image
    :return: reprojected coordinate array
    """
    reprojected_points = [
        lon_lat_to_pixel(point[0], point[1], geo_transform) for point in coords
    ]
    reprojected_points = [
        (int(point[0] / resize_factor), int(point[1] / resize_factor))
        for point in reprojected_points
    ]
    return reprojected_points


def overlay_polygon(
    img: np.ndarray,
    polygon: Polygon,
    color: Tuple[int, int, int, int] = (0, 255, 0, 0),
    opacity: float = 0.2
) -> np.ndarray:
    """
    Takes an input image and a polygon (which must exist in pixel coordinate space)
    and returns an image with the polygon transparently overlaid
    NOTE: only utilizes the exterior ring of the provided polygon
    :param img: The image on which the transparent polygon will be overlaid
    :param polygon: The shapely Polygon to overlay
    :param color: The color to make the polygon
    :param opacity: The alpha level of the polygon (ignore the alpha channel in color!!)
    :return: a modified image as an np.ndarray
    """
    polygon_coords = polygon.exterior.coords
    # needed for opencv /shrug
    polygon_array = np.array(polygon_coords, dtype=np.int32).reshape(-1, 1, 2)
    overlay = np.zeros((img.shape[0], img.shape[1]), dtype=np.ubyte)
    overlay = cv2.cvtColor(overlay, cv2.COLOR_GRAY2RGB)
    result = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    overlay = cv2.fillConvexPoly(overlay, polygon_array, color, lineType=cv2.LINE_AA)
    result = cv2.addWeighted(overlay, opacity, result, 1 - opacity, 0, result)

    return result


@jit(nopython=True)
def generate_triangle_mesh(elevation_data: np.ndarray, reach: int = 10) -> Tuple[np.array, np.array, np.array]:
    """
    Generates a triangle mesh from elevation data using a very simple and non-optimal but extremely fast procedure
    Produces a 3D vertex list as a Tuple of np.array
    :param elevation_data:
    :param reach: The resulting mesh will be centered on the origin (0,0,0), reach determines how wide in the (x and y)
    dimension the mesh will go. Does not have an effect on height (z).
    :return: a Tuple of vertex arrays (Tuple[X, Y, Z])
    """
    shape_x = elevation_data.shape[0]
    shape_y = elevation_data.shape[1]
    x = np.linspace(-reach, reach, int(shape_x))
    y = np.linspace(-reach, reach, int(shape_y))

    display_e_matrix = elevation_data / 1000
    new_x = np.zeros(x.shape[0] * y.shape[0] * 6)
    new_y = np.zeros(x.shape[0] * y.shape[0] * 6)
    new_elev = np.zeros(x.shape[0] * y.shape[0] * 6)
    idx = 0
    for i in range(1, x.shape[0] - 1):
        for j in range(1, y.shape[0] - 1):
            new_x[idx] = x[i-1]
            new_y[idx] = y[j-1]
            new_elev[idx] = display_e_matrix[i-1, j-1]
            new_x[idx + 1] = x[i]
            new_y[idx + 1] = y[j-1]
            new_elev[idx + 1] = display_e_matrix[i, j-1]
            new_x[idx + 2] = x[i]
            new_y[idx + 2] = y[j]
            new_elev[idx + 2] = display_e_matrix[i, j]
            new_x[idx + 3] = x[i-1]
            new_y[idx + 3] = y[j]
            new_elev[idx + 3] = display_e_matrix[i-1, j]
            new_x[idx + 4] = x[i-1]
            new_y[idx + 4] = y[j-1]
            new_elev[idx + 4] = display_e_matrix[i-1, j-1]
            new_x[idx + 5] = x[i]
            new_y[idx + 5] = y[j]
            new_elev[idx + 5] = display_e_matrix[i, j]
            idx = idx + 6
    return new_x, new_y, new_elev
